# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helper tests."""

from urllib.parse import quote_plus

import pytest

from invenio_app.helpers import get_safe_redirect_target


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            "https://example.org/search?page=1&q=&keywords=taxonomy&keywords=animali",
            "/search?page=1&q=&keywords=taxonomy&keywords=animali",
        ),
        ("/search?page=1&size=20", "/search?page=1&size=20"),
        ("https://localhost/search?page=1", "https://localhost/search?page=1"),
    ],
)
def test_get_safe_redirect_target(app, test_input, expected):
    """Test that only "localhost" is a trusted absolute redirect target."""
    with app.test_request_context("/?next={0}".format(quote_plus(test_input))):
        assert get_safe_redirect_target() == expected



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C) 2022-2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import sys
import types
from collections import namedtuple
from copy import deepcopy

import pytest
from flask import Flask, g
from flask_limiter import Limiter

from invenio_app.config import APP_DEFAULT_SECURE_HEADERS, set_rate_limit
from invenio_app.ext import useragent_and_ip_limit_key
from invenio_app.helpers import obj_or_import_string


@pytest.fixture()
def base_app():
    """Flask application fixture."""
    app_ = Flask("testapp")
    app_.config.update(
        SECRET_KEY="SECRET_KEY",
        TESTING=True,
    )

    app_.config["APP_DEFAULT_SECURE_HEADERS"] = deepcopy(APP_DEFAULT_SECURE_HEADERS)
    app_.config["APP_DEFAULT_SECURE_HEADERS"]["force_https"] = False

    @app_.route("/requestid")
    def requestid():
        from flask import g  # Prevent pytest problems

        return g.request_id if g and hasattr(g, "request_id") else ""

    @app_.route("/limited_rate")
    def limited_rate():
        return "test"

    @app_.route("/unlimited_rate")
    def unlimited_rate():
        return "test"

    return app_


@pytest.fixture()
def app_with_no_limiter(base_app):
    """Flask application fixture without limiter registered."""
    with base_app.app_context():
        yield base_app


@pytest.fixture()
def app(base_app):
    """Flask application fixture."""
    base_app.config.update(
        TRUSTED_HOSTS=["localhost"],
        RATELIMIT_APPLICATION=set_rate_limit,
        RATELIMIT_GUEST_USER="2 per second",
        RATELIMIT_AUTHENTICATED_USER="5 per second",
        RATELIMIT_PER_ENDPOINT={"unlimited_rate": "200 per second"},
        RATELIMIT_HEADERS_ENABLED=True,
    )
    Limiter(
        base_app,
        key_func=obj_or_import_string(
            base_app.config.get("RATELIMIT_KEY_FUNC"),
            default=useragent_and_ip_limit_key,
        ),
    )
    with base_app.app_context():
        yield base_app


@pytest.fixture()
def wsgi_apps():
    """Wsgi app fixture."""
    from invenio_base.app import create_app_factory
    from invenio_base.wsgi import create_wsgi_factory, wsgi_proxyfix

    def _config(app, **kwargs):
        app.config.update(
            SECRET_KEY="SECRET_KEY",
            TESTING=True,
        )
        app.config["APP_DEFAULT_SECURE_HEADERS"] = deepcopy(APP_DEFAULT_SECURE_HEADERS)
        app.config["APP_DEFAULT_SECURE_HEADERS"]["force_https"] = False

    # API
    create_api = create_app_factory(
        "invenio",
        config_loader=_config,
        wsgi_factory=wsgi_proxyfix(),
    )
    # UI
    create_ui = create_app_factory(
        "invenio",
        config_loader=_config,
        wsgi_factory=wsgi_proxyfix(),
    )
    # Combined
    create_app = create_app_factory(
        "invenio",
        config_loader=_config,
        wsgi_factory=wsgi_proxyfix(create_wsgi_factory({"/api": create_api})),
    )
    return create_app, create_ui, create_api


@pytest.fixture()
def push_rate_limit_to_context():
    """Push a custom rate limit to the Flask global context."""
    custom_rate_limit = "10 per second"
    setattr(g, "user_rate_limit", custom_rate_limit)
    return custom_rate_limit



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Celery application for Invenio flavours."""

from flask_celeryext import create_celery_app

from .factory import create_ui

celery = create_celery_app(
    create_ui(
        SENTRY_TRANSPORT="raven.transport.http.HTTPTransport",
        RATELIMIT_ENABLED=False,
    )
)
"""Celery application for Invenio.

Overrides SENTRY_TRANSPORT wih synchronous HTTP transport since Celery does not
deal nicely with the default threaded transport.
"""

# Trigger an app log message upon import. This makes Sentry logging
# work with `get_task_logger(__name__)`.
celery.flask_app.logger.info("Created Celery app")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Celery test."""


def test_celery():
    """Test celery application."""
    from invenio_app.celery import celery

    celery.loader.import_default_modules()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from collections import namedtuple
from unittest.mock import patch

from flask import current_app


def exhaust_and_test_rate_limit_per_second(client, rate_to_exhaust):
    """Helper function to test,per second,rate limits."""
    available_rate = int(rate_to_exhaust.split("per")[0])
    for x in range(0, available_rate):
        assert 200 == client.get("/limited_rate").status_code
    assert 429 == client.get("/limited_rate").status_code


def test_limiter(app):
    """Test the Flask limiter function."""
    with app.test_client() as client:
        exhaust_and_test_rate_limit_per_second(
            client, current_app.config["RATELIMIT_GUEST_USER"]
        )
        for x in range(0, 2):
            assert 200 == client.get("/unlimited_rate").status_code
        assert 200 == client.get("/unlimited_rate").status_code


FakeUser = namedtuple("User", ["is_authenticated"])


@patch("invenio_app.limiter.current_user", FakeUser(is_authenticated=True))
def test_limiter_for_authenticated_user(app):
    """Test the Flask limiter function."""
    with app.test_client() as client:
        exhaust_and_test_rate_limit_per_second(
            client, current_app.config["RATELIMIT_AUTHENTICATED_USER"]
        )
        assert 200 == client.get("/unlimited_rate").status_code


@patch("invenio_app.limiter.current_user", FakeUser(is_authenticated=True))
def test_limiter_for_privileged_user(app, push_rate_limit_to_context):
    """Test the Flask limiter function."""
    with app.test_client() as client:
        exhaust_and_test_rate_limit_per_second(client, push_rate_limit_to_context)

        assert 200 == client.get("/unlimited_rate").status_code



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from werkzeug.test import EnvironBuilder


def wsgi_output(application):
    data = {}

    def start_response(status, headers):
        data["status"] = status
        data["headers"] = headers

    data["output"] = application(
        EnvironBuilder(path="/", method="GET").get_environ(), start_response
    )
    return data


def test_wsgi(wsgi_apps):
    """Test wsgi."""
    create_app, create_ui, create_api = wsgi_apps

    res = wsgi_output(create_app())
    assert res["status"] == "404 NOT FOUND"

    res = wsgi_output(create_ui())
    assert res["status"] == "404 NOT FOUND"

    res = wsgi_output(create_api())
    assert res["status"] == "404 NOT FOUND"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Flask Limiter functions."""

from flask import current_app, g, request

try:
    from flask_login import current_user
except ModuleNotFoundError:
    current_user = None


def useragent_and_ip_limit_key():
    """Create key for the rate limiting."""
    return str(request.user_agent) + request.remote_addr


def set_rate_limit():
    r"""Set rates for Flask limiter.

    :return: a rate limit string with the Flask-Limiter format.

    For more information regarding the format you can see `here
    <https://flask-limiter.readthedocs
    .io/en/stable/#rate-limit-string-notation>`_.

    The order in which the rate will be evaluated is the following:

    1)Initially the endpoint is going to be evaluated against the whitelisted
    ones. If it has been marked as whitelisted then the custom limit for this
    endpoint will be the one to be returned.

    2)If the endpoint is not whitelisted and the flask-security package is
    installed, it will evaluate if the user is logged in and if this is the
    case it will also check if there is an explicitly defined rate limit for
    them by comparing theire ID with the ones present in the
    ```RATELIMIT_PER_USER``` mapping.
    If it is present then the custom rate limit value will be returned,
    otherwise the one returned will be the ```RATELIMIT_AUTHENTICATED_USER```

    3)Finally if none of the above is our case then the
    ```RATELIMIT_GUEST_USER``` will be the one to be returned.
    """
    endpoint_limits = current_app.config.get("RATELIMIT_PER_ENDPOINT", {})
    if request.endpoint in endpoint_limits:
        # Case of whitelisted endpoint.
        return endpoint_limits[request.endpoint]

    if current_user and current_user.is_authenticated:
        return g.get(
            "user_rate_limit", current_app.config["RATELIMIT_AUTHENTICATED_USER"]
        )
    return current_app.config["RATELIMIT_GUEST_USER"]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2019 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from flask import request, url_for

from invenio_app.factory import create_ui


def test_version():
    """Test version import."""
    from invenio_app import __version__

    assert __version__


def test_config_loader():
    """Test config loader."""
    app = create_ui()
    assert app.jinja_env.cache_size == 1000


def test_trusted_hosts():
    """Test trusted hosts configuration."""
    app = create_ui(
        TRUSTED_HOSTS=["example.org", "www.example.org"],
        APP_ENABLE_SECURE_HEADERS=False,
        RATELIMIT_ENABLED=False,
    )

    @app.route("/host")
    def index_host():
        return request.host

    @app.route("/url-for")
    def index_url():
        return url_for("index_url", _external=True)

    with app.test_client() as client:
        for u in ["/host", "/url-for"]:
            res = client.get(u, headers={"Host": "attacker.org"})
            assert res.status_code == 400

            res = client.get(u, headers={"Host": "example.org"})
            assert res.status_code == 200

            res = client.get(u, headers={"Host": "www.example.org"})
            assert res.status_code == 200


# There used to be a test here checking if the X-Forwarded-Host
#  is checked as well. It was deleted because from werkzeug 0.15.0 onwards
#  this check is not supported and therefore it caused the test to fail.
#  New way of dealing with proxies in werkzeug is `ProxyFix
#  <https://werkzeug.palletsprojects.com/en/0.15.x/middleware/proxy_fix`_.
#  The change in werkzeug mentioned above is visible in this
# `PR <https://github.com/pallets/werkzeug/pull/1303/files>`_.



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2019 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helpers."""

import os

from flask import current_app, request
from jinja2 import BaseLoader, TemplateNotFound
from uritools import uricompose, urisplit
from werkzeug.utils import cached_property, import_string


def get_safe_redirect_target(arg="next", _target=None):
    """Get URL to redirect to and ensure that it is local.

    :param arg: URL argument.
    :returns: The redirect target or ``None``.
    """
    for target in _target, request.args.get(arg), request.referrer:
        redirect = safe_redirect(target)
        if redirect:
            return redirect
    return None


def safe_redirect(target):
    """Ensure redirect is a local redirect."""
    if target:
        redirect_uri = urisplit(target)
        trusted_hosts = current_app.config.get("TRUSTED_HOSTS", [])
        if redirect_uri.host in trusted_hosts:
            return target
        elif redirect_uri.path:
            return uricompose(
                path=redirect_uri.path,
                query=redirect_uri.query,
                fragment=redirect_uri.fragment,
            )
    return None


class ThemeJinjaLoader(BaseLoader):
    """Prefix template loader.

    This loader acts as a wrapper for any type of Jinja loader. Before doing a
    template lookup, the loader sequentially applies prefixes to the template
    name, until a template source is found.

    The prefixes are defined via the ``APP_THEME`` configuration variable.
    """

    def __init__(self, app, loader):
        """Initialize loader.

        :param app: Flask application.
        :param loader: Jinja loader to be wrapped.
        """
        self.app = app
        self.loader = loader

    @cached_property
    def prefixes(self):
        """Return the active prefixes to be used for template lookup."""
        theme = self.app.config.get("APP_THEME", [])
        if isinstance(theme, str):
            theme = [theme]
        return theme

    def _prefixed_templates(self, name):
        template_names = [os.path.join(p, name) for p in self.prefixes]
        return template_names + [name]

    def get_source(self, environment, template):
        """Get the template source, filename and reload helper."""
        for tpl in self._prefixed_templates(template):
            try:
                return self.loader.get_source(environment, tpl)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(template)

    def load(self, environment, name, globals=None):
        """Loads a template."""
        for tpl in self._prefixed_templates(name):
            try:
                return self.loader.load(environment, tpl, globals)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(name)

    def list_templates(self):
        """List all availbale templates."""
        return self.loader.list_templates()


def obj_or_import_string(value, default=None):
    """Import string or return object.

    :params value: Import path or class object to instantiate.
    :params default: Default object to return if the import fails.
    :returns: The imported object.
    """
    if isinstance(value, str):
        return import_string(value)
    elif value:
        return value
    return default



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C) 2023-2025 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Flask application factories for Invenio flavours."""

import os
import sys

from invenio_base.app import create_app_factory
from invenio_base.urls import create_invenio_apps_urls_builder_factory
from invenio_base.wsgi import create_wsgi_factory, wsgi_proxyfix
from invenio_cache import BytecodeCache
from invenio_config import create_config_loader

env_prefix = "INVENIO"

invenio_config_loader = create_config_loader(config=None, env_prefix=env_prefix)


def instance_path():
    """Instance path for Invenio.

    Defaults to ``<env_prefix>_INSTANCE_PATH``
     or if environment variable is not set ``<sys.prefix>/var/instance``.
    """
    return os.getenv(env_prefix + "_INSTANCE_PATH") or os.path.join(
        sys.prefix, "var", "instance"
    )


def static_folder():
    """Static folder path.

    Defaults to ``<env_prefix>_STATIC_FOLDER``
    or if environment variable is not set ``<sys.prefix>/var/instance/static``.
    """
    return os.getenv(env_prefix + "_STATIC_FOLDER") or os.path.join(
        instance_path(), "static"
    )


def static_url_path():
    """Static url path.

    Defaults to ``<env_prefix>_STATIC_URL_PATH``
    or if environment variable is not set ``/static``.
    """
    return os.getenv(env_prefix + "_STATIC_URL_PATH") or "/static"


def config_loader(app, **kwargs_config):
    """Configuration loader."""
    invenio_config_loader(app, **kwargs_config)

    if app.config.get("APP_ALLOWED_HOSTS", None):
        app.config["TRUSTED_HOSTS"] = app.config["APP_ALLOWED_HOSTS"]
        app.logger.warning(
            "APP_ALLOWED_HOSTS is deprecated and has been replaced by TRUSTED_HOSTS."
        )

    # TODO: Move this to invenio_cache.ext.InvenioCache.init_app
    app.jinja_env.cache_size = 1000
    app.jinja_env.bytecode_cache = BytecodeCache(app)


def app_class():
    """Create Flask application class.

    Invenio-Files-REST needs to patch the Werkzeug form parsing in order to
    support streaming large file uploads. This is done by subclassing the Flask
    application class.
    """
    try:
        from invenio_files_rest.app import Flask
    except ImportError:
        from flask import Flask

    return Flask


create_api = create_app_factory(
    "invenio",
    config_loader=config_loader,
    blueprint_entry_points=["invenio_base.api_blueprints"],
    extension_entry_points=["invenio_base.api_apps"],
    converter_entry_points=["invenio_base.api_converters"],
    finalize_app_entry_points=["invenio_base.api_finalize_app"],
    wsgi_factory=wsgi_proxyfix(),
    instance_path=instance_path,
    root_path=instance_path,
    app_class=app_class(),
    urls_builder_factory=create_invenio_apps_urls_builder_factory(
        "SITE_API_URL",
        "SITE_UI_URL",
        {
            "blueprints": ["invenio_base.blueprints"],
            "converters": ["invenio_base.converters"],
        },
    ),
)
"""Flask application factory for Invenio REST API."""

create_ui = create_app_factory(
    "invenio",
    config_loader=config_loader,
    blueprint_entry_points=["invenio_base.blueprints"],
    extension_entry_points=["invenio_base.apps"],
    converter_entry_points=["invenio_base.converters"],
    finalize_app_entry_points=["invenio_base.finalize_app"],
    wsgi_factory=wsgi_proxyfix(),
    instance_path=instance_path,
    static_folder=static_folder,
    root_path=instance_path,
    static_url_path=static_url_path(),
    app_class=app_class(),
    urls_builder_factory=create_invenio_apps_urls_builder_factory(
        "SITE_UI_URL",
        "SITE_API_URL",
        {
            "blueprints": ["invenio_base.api_blueprints"],
            "converters": ["invenio_base.api_converters"],
        },
    ),
)
"""Flask application factory for Invenio UI."""

create_app = create_app_factory(
    "invenio",
    config_loader=config_loader,
    blueprint_entry_points=["invenio_base.blueprints"],
    extension_entry_points=["invenio_base.apps"],
    converter_entry_points=["invenio_base.converters"],
    finalize_app_entry_points=["invenio_base.finalize_app"],
    wsgi_factory=wsgi_proxyfix(create_wsgi_factory({"/api": create_api})),
    instance_path=instance_path,
    static_folder=static_folder,
    root_path=instance_path,
    static_url_path=static_url_path(),
    app_class=app_class(),
    urls_builder_factory=create_invenio_apps_urls_builder_factory(
        "SITE_UI_URL",
        "SITE_API_URL",
        {
            "blueprints": ["invenio_base.api_blueprints"],
            "converters": ["invenio_base.api_converters"],
        },
    ),
)
"""Flask application factory for combined UI + REST API.

REST API is mounted under ``/api``.
"""
